#include <windows.h>
#include <atlbase.h>
#include <atlstr.h>
#include <atlpath.h>

#include <boost/scoped_array.hpp>

//#undef OutputDebugString
//#define OutputDebugString(x)
//#define OutputDebugStringA(x)
//#define OutputDebugStringW(x)

#include "nazovb.h"

#undef _WINDLL
//#include <TTAlib.h>
#include "PngReader.h"
#include "OVReader.h"
#include "MoviePlayer.h"

#include "ProfileIO.h"

#include "resource.h"

#if 1
#include <stdarg.h>

void logprintf(const char *format, ...)
{
	va_list vl;
	va_start(vl, format);

	char s[512];
	vsprintf(s, format, vl);
	OutputDebugString(s);
}
#else
#define logprintf(...)
#endif


static LPCSTR s_MovExts[] = {
	".avi", ".flv", ".mpg", ".mpeg", ".mp4", ".divx", ".mov",
	NULL
};
static LPCSTR s_BGAExts[] = {
	".bmp", ".rle", ".jpg", ".png", ".gif",
	NULL
};
static LPCSTR s_SoundExts[] = {
	".wav", ".mp3", ".ogg", ".tta",
	NULL
};

static HMODULE g_hLibModule;

// ---------------------------------------------------------------------------

inline void WriteCall32(DWORD_PTR addr, void *call)
{
	*reinterpret_cast<BYTE *>(addr + 0) = 0xE8;
	*reinterpret_cast<long *>(addr + 1) =
		reinterpret_cast<DWORD_PTR>(call) - addr - 5;
}

inline void WriteJump32(DWORD_PTR addr, void *call)
{
	*reinterpret_cast<BYTE *>(addr + 0) = 0xE9;
	*reinterpret_cast<long *>(addr + 1) =
		reinterpret_cast<DWORD_PTR>(call) - addr - 5;
}

inline void WritePush32(DWORD_PTR addr, DWORD_PTR imm)
{
	*reinterpret_cast<BYTE *>(addr + 0) = 0x68;
	*reinterpret_cast<DWORD_PTR *>(addr + 1) = imm;
}

inline void WriteJump8(DWORD_PTR addr, void *call)
{
	*reinterpret_cast<BYTE *>(addr + 0) = 0xEB;
	*reinterpret_cast<BYTE *>(addr + 1) = 
		static_cast<BYTE>(reinterpret_cast<DWORD_PTR>(call) - addr - 2);
}

// ---------------------------------------------------------------------------

bool FileExists(const ATL::CString &path)
{
	DWORD attr = GetFileAttributes(path);
	return attr != 0xFFFFFFFF && !(attr & FILE_ATTRIBUTE_DIRECTORY);
}

bool FixFileExtension(ATL::CString &path, LPCSTR ext[])
{
	ATL::CPathT<ATL::CString> tmpath(path);
	do {
		if(!tmpath.RenameExtension(*ext))
			return false;
		if(FileExists(tmpath)) {
			path.operator=(tmpath.m_strPath);
			return true;
		}
	} while(*++ext);
	return false;
}

// ---------------------------------------------------------------------------

HRESULT _stdcall LoadPicturePng(
	VARIANT FileName,
	IPictureDisp** retval)
{
	if (FileName.vt != VT_BSTR) {
		return E_FAIL;
	}

	pngfile pf;
	if(!pf.open(vbstr_t(FileName.bstrVal, true).ToAnsi())) {
		return E_FAIL; 
	}

	BITMAPINFO bi;
	ZeroMemory(&bi, sizeof(BITMAPINFO));
	bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bi.bmiHeader.biBitCount = 32;
	bi.bmiHeader.biHeight = pf.height();
	bi.bmiHeader.biWidth = pf.width();
	bi.bmiHeader.biPlanes = 1;
	bi.bmiHeader.biCompression = BI_RGB;

	void *buff;
	HBITMAP bmp = CreateDIBSection(NULL, &bi, DIB_RGB_COLORS, &buff, NULL, 0);
	if(!bmp) return E_FAIL;
	pf.readall(static_cast<png_bytep>(buff));

	PICTDESC pd;
	pd.cbSizeofstruct = sizeof(pd);
	pd.picType = PICTYPE_BITMAP;
	pd.bmp.hbitmap = bmp;
	pd.bmp.hpal = NULL;
	HRESULT hr = OleCreatePictureIndirect(&pd, IID_IPictureDisp, TRUE, reinterpret_cast<void **>(retval));
	if(FAILED(hr)) {
		DeleteObject(bmp);
	}

	return hr;
}

HRESULT _stdcall LoadPictureHook(
	void *This,
    /* [in, optional] */ VARIANT FileName, 
    /* [in, optional] */ VARIANT Size, 
    /* [in, optional] */ VARIANT ColorDepth, 
    /* [in, optional] */ VARIANT X, 
    /* [in, optional] */ VARIANT Y, 
    /* [out, retval] */ IPictureDisp** retval)
{
	HRESULT hr = LoadPicturePng(FileName, retval);
	if(FAILED(hr)) {
		hr = mGlobal->LoadPicture(FileName, Size, ColorDepth, X, Y, retval);
	}
	return hr;
}

VARIANT_BOOL _stdcall LoadNullBGA(
	const BSTR *pFileName,
	DirectDrawSurface7 **ppRetVal,
	VBDDSURFACEDESC2 *pSurfDesc)
{
	ZeroMemory(pSurfDesc, sizeof(VBDDSURFACEDESC2));
	pSurfDesc->lFlags = (CONST_DDSURFACEDESCFLAGS)(vbDDSD_CAPS | vbDDSD_HEIGHT | vbDDSD_WIDTH);
	pSurfDesc->lWidth = pSurfDesc->lHeight = 16;
	pSurfDesc->ddsCaps.lCaps = vbDDSCAPS_OFFSCREENPLAIN;
	if(FAILED(mDDraw->CreateSurface(pSurfDesc, ppRetVal)))
		return FALSE;
	RECT rect = { 0, 0, 16, 16 };
	long status;
	(*ppRetVal)->BltColorFill(&rect, 0, &status);
	return VARIANT_TRUE;
}

VARIANT_BOOL __declspec(naked) _stdcall CreateSurfaceFromFileStub(
	const BSTR *pFileName,
	DirectDrawSurface7 **ppRetVal,
	VBDDSURFACEDESC2 *pSurfDesc)
{
	__asm {
		push ebp
		mov ebp, esp
		sub esp, 18h
		mov eax, 0x004C7A36
		jmp eax
	}
}

VARIANT_BOOL _stdcall CreateSurfaceFromFileCustomHook(
	const BSTR *pFileName,
	DirectDrawSurface7 **ppRetVal,
	VBDDSURFACEDESC2 *pSurfDesc)
{
	ATL::CString filename(vbstr_t(*pFileName, true).ToAnsi());
	if(!FixFileExtension(filename, s_BGAExts)) 
		return FALSE;

	return CreateSurfaceFromFileStub(&vbstr_t(filename).GetBSTR(), ppRetVal, pSurfDesc);
}

HRESULT _stdcall CreateSurfaceFromFileCustomHook2(
	const BSTR pFileName,
	VBDDSURFACEDESC2 *pSurfDesc,
	DirectDrawSurface7 **ppRetVal)
{
	ATL::CString filename(vbstr_t(pFileName, true).ToAnsi());
	if(!FixFileExtension(filename, s_BGAExts)) return E_FAIL;

	return CreateSurfaceFromFileStub(&vbstr_t(filename).GetBSTR(), ppRetVal, pSurfDesc)
		? NOERROR
		: E_FAIL;
}

// ---------------------------------------------------------------------------

HRESULT WINAPI CreateSoundBufferFromOgg(
	DirectSound8 *This,
	/* [in] */ LPCSTR filename, 
	/* [in, out] */ VBDSBUFFERDESC* bufferDesc, 
	/* [out, retval] */ DirectSoundSecondaryBuffer8** soundBuffer)
{
	ovfile vf;
	if(!vf.open(filename, NULL, 0)) return E_FAIL;

	vorbis_info *vi = vf.info();
	ogg_int64_t total = vf.pcm_total();

	bufferDesc->lSize = sizeof(*bufferDesc);
	bufferDesc->lFlags = static_cast<CONST_DSBCAPSFLAGS>
		(DSBCAPS_STATIC | DSBCAPS_GLOBALFOCUS | DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLVOLUME);
	bufferDesc->guid3DAlgorithm = NULL;
	bufferDesc->lBufferBytes = static_cast<long>(total * vi->channels * 2);
	bufferDesc->lReserved = 0;
	bufferDesc->fxFormat.nChannels = vi->channels;
	bufferDesc->fxFormat.nSamplesPerSec = vi->rate;
	bufferDesc->fxFormat.wBitsPerSample = 16;
	bufferDesc->fxFormat.wFormatTag = WAVE_FORMAT_PCM;
	bufferDesc->fxFormat.nBlockAlign = bufferDesc->fxFormat.nChannels * 2;
	bufferDesc->fxFormat.nAvgBytesPerSec = bufferDesc->fxFormat.nSamplesPerSec * bufferDesc->fxFormat.nBlockAlign;

	CComPtr<DirectSoundSecondaryBuffer8> dsb;
	HRESULT hr = This->CreateSoundBuffer(bufferDesc, &dsb);
	if(FAILED(hr)) return hr;

	try {
		boost::scoped_array<char> buf
			(new char[bufferDesc->lBufferBytes]);

		long len = vf.read(buf.get(), bufferDesc->lBufferBytes);
		if(len == 0) return E_FAIL;

		hr = dsb->WriteBuffer(0, len, buf.get(), vbDSBLOCK_ENTIREBUFFER);
		if(FAILED(hr)) return hr;

		*soundBuffer = dsb.Detach();
	} catch(std::bad_alloc &) {
		return E_OUTOFMEMORY;
	} catch(...) {
		return E_FAIL;
	}

	return NOERROR;
}

//HRESULT WINAPI CreateSoundBufferFromTTA(
//	DirectSound8 *This,
//	LPCSTR filename, 
//	DSBUFFERDESC* bufferDesc, 
//	DirectSoundSecondaryBuffer8** soundBuffer)
//{
//	try {
//		TTALib::TTADecoder decoder(filename);
//
//		bufferDesc->lSize = sizeof(*bufferDesc);
//		bufferDesc->lFlags = static_cast<CONST_DSBCAPSFLAGS>
//			(DSBCAPS_STATIC | DSBCAPS_GLOBALFOCUS | DSBCAPS_CTRLFREQUENCY | DSBCAPS_CTRLVOLUME);
//		bufferDesc->fxFormat.nChannels = decoder.GetNumChannels();
//		bufferDesc->fxFormat.nSamplesPerSec = decoder.GetSampleRate();
//		bufferDesc->fxFormat.wBitsPerSample = decoder.GetBitsPerSample();
//		bufferDesc->fxFormat.wFormatTag = decoder.GetAudioFormat();
//		bufferDesc->fxFormat.nBlockAlign = bufferDesc->fxFormat.nChannels * (bufferDesc->fxFormat.wBitsPerSample / 8);
//		bufferDesc->fxFormat.nAvgBytesPerSec = bufferDesc->fxFormat.nSamplesPerSec * bufferDesc->fxFormat.nBlockAlign;
//		bufferDesc->guid3DAlgorithm = NULL;
//		bufferDesc->lBufferBytes = decoder.GetDataLength() * bufferDesc->fxFormat.nBlockAlign;
//		bufferDesc->lReserved = 0;
//
//		CComPtr<DirectSoundSecondaryBuffer8> dsb;
//		HRESULT hr = This->CreateSoundBuffer(bufferDesc, &dsb);
//		if(FAILED(hr)) return hr;
//
//		long *block, len;
//		while(len = decoder.GetBlock(&block)) {
//
//			boost::scoped_array<char> buf
//				(new char[bufferDesc->lBufferBytes]);
//
//			OutputDebugString("Writing...");
//			hr = dsb->WriteBuffer(
//				0,
//				len * bufferDesc->fxFormat.nBlockAlign,
//				block,
//				DSBLOCK_FROMWRITECURSOR);
//			if(FAILED(hr)) {
//				OutputDebugString("Failed Writing");
//				return hr;
//			}
//		}
//		*soundBuffer = dsb.Detach();
//	} catch(std::bad_alloc &) {
//		return E_OUTOFMEMORY;
//	} catch(...) {
//		return E_FAIL;
//	}
//	return NOERROR;
//}

//extern "C"
//void * _ReturnAddress(void);
//#pragma intrinsic(_ReturnAddress)
//
//void __stdcall DSBuff_LoadSoundHook(
//	void *This,
//	const BSTR *pFileName,
//	VARIANT_BOOL *pRetVal)
//{
//	typedef void (__stdcall *DSBuff_LoadSoundType)(void *, const BSTR *, VARIANT_BOOL *);
//	logprintf("%p", _ReturnAddress());
//	reinterpret_cast<DSBuff_LoadSoundType>(0x004DEA30)(This, pFileName, pRetVal);
//}


void WINAPI SimpleLoadBMS_Hook(BSTR FileName)
{
	typedef void (__stdcall *SimpleLoadBMS_Type)(BSTR *pFileName);
	typedef void (__stdcall *DrawScene_Type)(int forceVSync, int forceNoSysStr, int forceDBlt);

	reinterpret_cast<SimpleLoadBMS_Type>(0x00492D40)(&FileName);
	while(GetAsyncKeyState(VK_RETURN) >= 0) {
		reinterpret_cast<DrawScene_Type>(0x004CA680)(0, -1, 0);
		rtcDoEvents();
	}
}

void __declspec(naked) DSBuff_LoadSoundEpilogueHook()
{
	__asm {
		mov word ptr [esi+0x6c], VARIANT_TRUE
		push 0x4DECD6
		ret
	}
}

DirectSoundSecondaryBuffer8* WINAPI LoadSoundHook(
	/* [in] */ const BSTR *pFileName, 
	/* [in, out] */ VBDSBUFFERDESC* pBufDesc)
{
	typedef DirectSoundSecondaryBuffer8 * (WINAPI *LoadSound_Type)(const BSTR *, VBDSBUFFERDESC *);

	ATL::CString filename(vbstr_t(*pFileName, true).ToAnsi());
	if(!FixFileExtension(filename, s_SoundExts)) return NULL;

	DirectSoundSecondaryBuffer8 *sndbuf = NULL;
	HRESULT hr = CreateSoundBufferFromOgg(mDSound, filename, pBufDesc, &sndbuf);
	if(SUCCEEDED(hr)) return sndbuf;
	//hr = CreateSoundBufferFromTTA(mDSound, filename, pBufDesc, &sndbuf);
	//if(SUCCEEDED(hr)) return sndbuf;

	return reinterpret_cast<LoadSound_Type>(0x004DA590)(pFileName, pBufDesc);
}

// ---------------------------------------------------------------------------

// movieplayer* movie_cache[128][36 * 36];
movieplayer* movie_array[36 * 36];
movieplayer* current_movie;
POINT form_origin;

LRESULT CALLBACK SubclassWndProcHook(HWND hwnd, UINT umsg, WPARAM wparam, LPARAM lparam)
{
	if(umsg == WM_MOVE) {
		RECT form_rect;
		GetClientRect(hwnd, &form_rect);
		form_origin.x = form_rect.left;
		form_origin.y = form_rect.top;
		ClientToScreen(hwnd, &form_origin);
	}
	return reinterpret_cast<WNDPROC>(0x00484C30)(hwnd, umsg, wparam, lparam);
}

void RenderBGAHook()
{
	RECT nullrect = {0, 0, 0, 0};
	if(
	mShowNormalBGA == -1 &&
	mBGAID != -1 &&
	movie_array[mBGAID]) {
		if(current_movie) {
			current_movie->SetVideoRect(&nullrect);
			current_movie->Stop();
			current_movie->SetPosition(0, 0);
		}
		current_movie = movie_array[mBGAID];
		current_movie->Play();
		mBGAID = -1;
		mShowNormalBGA = VARIANT_FALSE;
	}
	if(current_movie) {
		if(current_movie->IsStoped()) {
			current_movie->SetVideoRect(&nullrect);
			current_movie->Stop();
			current_movie->SetPosition(0, 0);
			current_movie = NULL;
		} else {
			RECT fixSkinBGARect = { 
				mSkinBGARect.left + form_origin.x,
				mSkinBGARect.top + form_origin.y,
				mSkinBGARect.left + mSkinBGARect.right + form_origin.x,
				mSkinBGARect.top + mSkinBGARect.bottom + form_origin.y
			};
			current_movie->SetVideoRect(
				mShowPoorBGA == VARIANT_TRUE ? 
				&nullrect : &fixSkinBGARect);
		}
	}

	reinterpret_cast<void (*)()>(0x00481950)();
}

void __stdcall UnloadBMSHook(VARIANT_BOOL *pRetVal)
{
	if(current_movie) {
		RECT nullrect = {0, 0, 0, 0};
		current_movie->SetVideoRect(&nullrect);
		current_movie->Stop();
		current_movie = NULL;
	}
	for(int j = 0; j < 36 * 36; ++j) {
		if(movie_array[j]) {
			delete movie_array[j];
			movie_array[j] = NULL;
		}
	}
	reinterpret_cast<void (__stdcall *)(VARIANT_BOOL *pRetVal)>(0x004BA1E0)(pRetVal);
}

void LoadMovie(int BMSID, const BSTR *pFileName)
{
	if(movie_array[BMSID]) {
		OutputDebugString("movie_array[BMSID]");
		return;
	}

	ATL::CAutoPtr<movieplayer> movie(new movieplayer());
	movie->SetFilePath(*pFileName);

	CComPtr<IUnknown> dd7;
	CComPtr<IUnknown> prim7;
	if(FAILED(mDDraw->InternalGetObject(&dd7))) {
		OutputDebugString("mDDraw->InternalGetObject");
		return;
	}
	if(FAILED(mPrimSurf->InternalGetObject(&prim7))) {
		OutputDebugString("mPrimSurf->InternalGetObject");
		return;
	}

	CComQIPtr<IDirectDraw, &IID_IDirectDraw> dd(dd7);
	if(!dd) {
		OutputDebugString("dd7->QueryInterface");
		return;
	}
	CComQIPtr<IDirectDrawSurface, &IID_IDirectDrawSurface> prim(prim7);
	if(!prim) {
		OutputDebugString("prim7->QueryInterface");
		return;
	}
	if(movie->BuildGraph(dd, prim)) {
		dd.Detach(); prim.Detach();
		movie_array[BMSID] = movie.Detach();
	} else {
		OutputDebugString("movie->BuildGraph");
	}
}

VARIANT_BOOL _stdcall LoadBGAHook(
    int BMSID,
	const BSTR *pFileName,
	DirectDrawSurface7 **ppRetVal,
	VBDDSURFACEDESC2 *pSurfDesc)
{
	ATL::CString filename(vbstr_t(*pFileName, true).ToAnsi());
	if(!FixFileExtension(filename, s_MovExts))
		return CreateSurfaceFromFileCustomHook(pFileName, ppRetVal, pSurfDesc);

	LoadMovie(BMSID, pFileName);
	return LoadNullBGA(pFileName, ppRetVal, pSurfDesc) == S_OK;
}

void __declspec(naked) LoadBGAHook_Stub()
{
	__asm {
		pop eax
		push dword ptr ds:[ebp + 8]
		push eax
		jmp LoadBGAHook
	}
}

// ---------------------------------------------------------------------------

void InitMainFrmHook()
{
	if(mWindowed) {
		if(!mMainFrm) {
			__vbaNew2(
				reinterpret_cast<void **>(0x00412790), 
				reinterpret_cast<IUnknown **>(&mMainFrm));
		}
		CProfileIO profile("\\_nbmfiles\\nbm_conf.ini");
		mMainFrm->Left(static_cast<float>(profile.ReadInteger("MainForm", "Left", 0)));
		mMainFrm->Top(static_cast<float>(profile.ReadInteger("MainForm", "Top", 0)));
	}
	reinterpret_cast<void (*)()>(0x00445470)();
}

// ---------------------------------------------------------------------------

void PlayLoopHook()
{
	HIMC imc;
	HWND hwnd;
	mMainFrm->hWnd(reinterpret_cast<long *>(&hwnd));

	imc = ImmAssociateContext(hwnd, NULL);

	// PlayLoop
	reinterpret_cast<void (*)()>(0x00434D80)();

	ImmAssociateContext(hwnd, imc);
}
// ---------------------------------------------------------------------------

//static unsigned long s_repseed;
//static unsigned long s_recseed;
//
//HRESULT WINAPI __vbaFileCloseHook1(short filenumber)
//{
//	__vbaGet4(4, &s_repseed, 0x59C, filenumber);
//	return __vbaFileClose(filenumber);
//}
//
//HRESULT WINAPI __vbaFileCloseHook2(short filenumber)
//{
//	__vbaPut4(4, &s_recseed, 0x59C, filenumber);
//	return __vbaFileClose(filenumber);
//}
//
//void __declspec(naked) ResetVBRandom()
//{
//	__asm {
//		sub esp, 50h
//		mov dword ptr [esp], VT_I2
//		mov dword ptr [esp+8], 0xFFFFFF80
//		push esp
//		call ds:[0x00401118]
//		fstp st(0)
//		push esp
//		call ds:[00401060]
//		add esp, 30h
//		ret
//	}
//}
//
//void WINAPI BMSRelocateObjHook(void *arg)
//{
//	ResetVBRandom();
//	VARIANT seed;
//	seed.vt = VT_I4;
//	seed.uintVal = mReplayMode ? s_repseed : (s_recseed = GetTickCount());
//	rtcRandomize(&seed);
//	rtcFreeVar(&seed);
//
//	reinterpret_cast<void (WINAPI *)(void *arg)>(0x004AD830)(arg);
//}

// ---------------------------------------------------------------------------

//struct NAZOBJ {
//	long   data00;
//	double frame;
//	long   data10;
//	long   id;
//};
//
//static struct {
//	int origid;
//	double frame;
//} s_lastSara[2];
//
//static const int s_pivot[] = { 1, 2, 3, 4, 5, 6, 8, 9 };
//
//bool s_specialopt;
//bool s_sarather;
//bool s_sarather_hyper;
//long s_sarather_hyper_count[2];
//long s_sarather_hyper_max[2];
//long s_sarather_pivot[2];
//extern "C"
//void * _AddressOfReturnAddress(void);
//#pragma intrinsic(_AddressOfReturnAddress)
//
//void RelocateObjHook()
//{
//	// 炴[
//	if(GetAsyncKeyState(VK_NUMPAD7) < 0) {
//		ZeroMemory(s_lastSara, sizeof(s_lastSara));
//		s_sarather = true;
//		s_sarather_hyper = false;
//		if(GetAsyncKeyState(VK_DIVIDE) < 0) {
//			s_sarather_pivot[0] = s_sarather_pivot[1] = 6;
//		} else {
//			s_sarather_pivot[0] =
//			s_sarather_pivot[1] = s_pivot[rand() % 8];
//			if(GetAsyncKeyState(VK_MULTIPLY)) {
//				s_sarather_hyper = true;
//				s_sarather_hyper_count[0] = 0;
//				s_sarather_hyper_count[1] = 0;
//				s_sarather_hyper_max[0] = ((rand() % 6) + 1);
//				s_sarather_hyper_max[1] = ((rand() % 6) + 1);
//			}
//		}
//	} else {
//		s_sarather = false;
//	}
//	s_specialopt = s_sarather;
//	reinterpret_cast<void (*)()>(0x004AD830)();
//}
//
//void __stdcall ShowResultSceneHook()
//{
//	if(s_specialopt) {
//		VARIANT_BOOL mIREnabled  = *reinterpret_cast<VARIANT_BOOL *>(0x00503710);
//		*reinterpret_cast<VARIANT_BOOL *>(0x00503710) = VARIANT_FALSE;
//		reinterpret_cast<void (*)()>(0x004BC0A0)();
//		*reinterpret_cast<VARIANT_BOOL *>(0x00503710) = mIREnabled;
//	} else {
//		((FARPROC)0x004BC0A0)();
//	}
//}
//
//void __stdcall WriteToNSDHook()
//{
//	if(!s_specialopt)
//		((FARPROC)0x004BBE40)();
//}
//
//void __stdcall SaratherHook(NAZOBJ *obj)
//{
//	if(s_sarather) {
//		long y = obj->id / 10 - 1;
//		long x = (y + 1) * 10;
//
//		if(x + 7 == obj->id) goto fini;
//
//		if(s_sarather_hyper &&
//		   obj->frame != s_lastSara[y].frame) {
//			if(++s_sarather_hyper_count[y] > s_sarather_hyper_max[y]) {
//				long pivot;
//				do { pivot = s_pivot[rand() % 8]; }
//				while(pivot == s_sarather_pivot[y]);
//				s_sarather_pivot[y] = pivot;
//				s_sarather_hyper_count[y] = 0;
//				s_sarather_hyper_max[y] = ((rand() % 6) + 1);
//			}
//		}
//
//		if(x + s_sarather_pivot[y] == obj->id) {
//			if(!s_lastSara[y].origid) {
//				s_lastSara[y].origid = obj->id;
//				s_lastSara[y].frame  = obj->frame;
//			} else if(
//			obj->frame == s_lastSara[y].frame &&
//			obj->id    != s_lastSara[y].origid)
//				obj->id = s_lastSara[y].origid;
//			goto fini;
//		}
//
//		if(s_lastSara[y].origid &&
//		   obj->frame == s_lastSara[y].frame) goto fini;
//
//		s_lastSara[y].origid = obj->id;
//		s_lastSara[y].frame  = obj->frame;
//		obj->id = x + s_sarather_pivot[y];
//	}
//fini:
//	if(
//	*reinterpret_cast<short *>(0x0050352E) ||
//	*reinterpret_cast<short *>(0x0050352C) ||
//	*reinterpret_cast<short *>(0x0050352A)) {
//		*reinterpret_cast<DWORD_PTR *>(_AddressOfReturnAddress()) = 0x004B03BA;
//	} else {
//		*reinterpret_cast<DWORD_PTR *>(_AddressOfReturnAddress()) = 0x004B0807;
//	}
//}

// ---------------------------------------------------------------------------

#define HANDLE_DLGMSG(hwnd, message, fn)    \
    case (message): return (SetDlgMsgResult(hwnd, message, \
		HANDLE_##message((hwnd), (wParam), (lParam), (fn))));


LRESULT ConfigureDlg_OnInitDialog(HWND hwndDlg, HWND hwndFirstCtl, LPARAM dwInitParam)
{
	CProfileIO profile("\\_nbmfiles\\nbm_conf.ini");
	CheckDlgButton(hDlg, IDC_CUSTOM_SINGLEPLAY, 
		profile.ReadInteger("Glasopal", "CustomSpecifiedBMSPlay", 0) ? BST_CHECKED: BST_UNCHECKED);
	//CheckDlgButton(hDlg, IDC_CUSTOM_OPTIONS, 
	//	profile.ReadInteger("Glasopal", "CustomOptions", 0) ? BST_CHECKED: BST_UNCHECKED);
	CheckDlgButton(hDlg, IDC_DISABLE_IME, 
		profile.ReadInteger("Glasopal", "DisableIME", 0) ? BST_CHECKED: BST_UNCHECKED);
	CheckDlgButton(hDlg, IDC_BGAMOVIE, 
		profile.ReadInteger("Glasopal", "BGAMovie", 0) ? BST_CHECKED: BST_UNCHECKED);
	CheckDlgButton(hDlg, IDC_WINDOW_MEMORIZE, 
		profile.ReadInteger("Glasopal", "WindowMemorize", 0) ? BST_CHECKED: BST_UNCHECKED);
	CheckDlgButton(hDlg, IDC_AUTO_PRIORITY,
		profile.ReadInteger("Glasopal", "AutoPriority", 0) ? BST_CHECKED: BST_UNCHECKED);
	CheckDlgButton(hDlg, IDC_MINIMUM_COMBO_EFFECT,
		profile.ReadInteger("Glasopal", "MinimumComboEffect", 0) ? BST_CHECKED: BST_UNCHECKED);
	return TRUE;
}

void ConfigureDlg_OnCommand(HWND hwndDlg, int id, HWND hwndCtl, UINT codeNotify)
{
	switch(LOWORD(wParam)) {
	case IDOK:
		{
			CProfileIO profile("\\_nbmfiles\\nbm_conf.ini");
			profile.WriteInteger("Glasopal", "CustomSpecifiedBMSPlay",
				IsDlgButtonChecked(hDlg, IDC_CUSTOM_SINGLEPLAY) == BST_CHECKED);
			//profile.WriteInteger("Glasopal", "CustomOptions",
			//	IsDlgButtonChecked(hDlg, IDC_CUSTOM_OPTIONS) == BST_CHECKED);
			profile.WriteInteger("Glasopal", "DisableIME",
				IsDlgButtonChecked(hDlg, IDC_DISABLE_IME) == BST_CHECKED);
			profile.WriteInteger("Glasopal", "BGAMovie",
				IsDlgButtonChecked(hDlg, IDC_BGAMOVIE) == BST_CHECKED);
			profile.WriteInteger("Glasopal", "WindowMemorize",
				IsDlgButtonChecked(hDlg, IDC_WINDOW_MEMORIZE) == BST_CHECKED);
			profile.WriteInteger("Glasopal", "AutoPriority",
				IsDlgButtonChecked(hDlg, IDC_AUTO_PRIORITY) == BST_CHECKED);
			profile.WriteInteger("Glasopal", "MinimumComboEffect",
				IsDlgButtonChecked(hDlg, IDC_MINIMUM_COMBO_EFFECT) == BST_CHECKED);
			MessageBox(hDlg, "܂", "Glasopal", MB_OK);
		}
	case IDCANCEL:
		EndDialog(hDlg, 0);
		return TRUE;
	}
}

BOOL CALLBACK ConfigureDlgProc(HWND hDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch(uMsg) {
	HANDLE_DLGMSG(hwndDlg, WM_INITDIALOG, ConfigureDlg_OnInitDialog)
	HANDLE_DLGMSG(hwndDlg, WM_COMMAND, ConfigureDlg_OnCommand)
	}
	return FALSE;
}

// ---------------------------------------------------------------------------


//static long x;
//HRESULT WINAPI vbsafdHook(SAFEARRAY *dst, SAFEARRAY *src, UUID *uuid)
//{
//
//	if(src->rgsabound[0].cElements == 0x240) {
//		char buf[128];
//		sprintf(buf, "d=%p e=%p", &dst->pvData, &src->rgsabound[0].cElements);
//		OutputDebugString(buf);
//	}
//	return reinterpret_cast<HRESULT (WINAPI *)(SAFEARRAY *dst, SAFEARRAY *src, UUID *uuid)>(x)(dst, src, uuid);
//	
//}


void APIENTRY OnMain()
{
	if(GetAsyncKeyState(VK_PAUSE)) {
		DialogBoxParamA(g_hLibModule, MAKEINTRESOURCE(IDD_CONFIGURE), NULL, ConfigureDlgProc, 0);
		return;
	}
	//// ΂񂲁[Ђ[
	//x = WriteMem<long>(0x660DA407, (long)vbsafdHook - 0x660DA40B) + 0x660DA40B;

	// ΂񂲁[Ђ[
	*reinterpret_cast<DWORD_PTR>(0x0043362B) = 0x502EA0;

	// 502B80(mCourseBGAArray)
	*reinterpret_cast<PINT32>(0x0041B8D4) = 0x534; 

	*reinterpret_cast<PINT32>(0x0047C1ED) = 0x534;
	*reinterpret_cast<PINT32>(0x0047C3A2) = 0x534;
	*reinterpret_cast<PINT32>(0x0047C9D1) = 0x534;
	*reinterpret_cast<PINT32>(0x0047CA6B) = 0x534;
	*reinterpret_cast<PINT32>(0x0047CC4F) = 0x534;
	*reinterpret_cast<PINT32>(0x0047CD3E) = 0x534;
	*reinterpret_cast<PINT32>(0x00480F9D) = 0x534;
	*reinterpret_cast<PINT32>(0x00481012) = 0x534;
	*reinterpret_cast<PINT32>(0x00481454) = 0x534;
	*reinterpret_cast<PINT32>(0x004814D5) = 0x534;
	
	// 502BA4(mCourseBGADescArray)
	*reinterpret_cast<PINT32>(0x0041B908) = 0x534;

	*reinterpret_cast<PINT32>(0x0047CE20) = 0x534;
	*reinterpret_cast<PINT32>(0x004810AD) = 0x534;
	*reinterpret_cast<PINT32>(0x00481582) = 0x534;

	// 502BC8(mCourseBGAPathArray)
	*reinterpret_cast<PINT32>(0x0041B938) = 0x534;

	*reinterpret_cast<PINT32>(0x0047C29D) = 0x534;
	*reinterpret_cast<PINT32>(0x004B44CA) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA61D) = 0x534;

	// 502BEC(mCourseBGALoaded)
	*reinterpret_cast<PINT32>(0x0041B96C) = 0x534;

	*reinterpret_cast<PINT32>(0x004387F1) = 0x534;
	*reinterpret_cast<PINT32>(0x00438CB8) = 0x534;
	*reinterpret_cast<PINT32>(0x004811B1) = 0x534;
	*reinterpret_cast<PINT32>(0x00481686) = 0x534;
	*reinterpret_cast<PINT32>(0x004B45D3) = 0x534;

	// 502C68(????????)
	*reinterpret_cast<PINT32>(0x0041BA24) = 0x534;

	// 502F0C(mSoundLoaded)
	*reinterpret_cast<PINT32>(0x0041BD2A) = 0x534; 

	*reinterpret_cast<PINT32>(0x00440034) = 0x534;
	*reinterpret_cast<PINT32>(0x004407A4) = 0x534;
	*reinterpret_cast<PINT32>(0x004B1AB4) = 0x534;
	*reinterpret_cast<PINT32>(0x004B38C1) = 0x534;
	*reinterpret_cast<PINT32>(0x004BAF66) = 0x534;

	// 502F28(mBGALoaded)
	*reinterpret_cast<PINT32>(0x0041BD56) = 0x534;

	*reinterpret_cast<PINT32>(0x0043881E) = 0x534;
	*reinterpret_cast<PINT32>(0x00438D09) = 0x534;
	*reinterpret_cast<PINT32>(0x00480F3B) = 0x534;
	*reinterpret_cast<PINT32>(0x00481133) = 0x534;
	*reinterpret_cast<PINT32>(0x004811F9) = 0x534;
	*reinterpret_cast<PINT32>(0x004813E6) = 0x534;
	*reinterpret_cast<PINT32>(0x00481608) = 0x534;
	*reinterpret_cast<PINT32>(0x004816CE) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA377) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA859) = 0x534;

	// 502F44(mBPMVector)
	*reinterpret_cast<PINT32>(0x0041BD82) = 0x534; 

	*reinterpret_cast<PINT32>(0x00439941) = 0x534;
	*reinterpret_cast<PINT32>(0x00439996) = 0x534; 
	*reinterpret_cast<PINT32>(0x004AEFD2) = 0x534;
	*reinterpret_cast<PINT32>(0x004B5050) = 0x534;

	// 502F60(mLoadedBGAPathArray)
	*reinterpret_cast<PINT32>(0x0041BDAE) = 0x534;

	*reinterpret_cast<PINT32>(0x004BA338) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA7CE) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA8CE) = 0x534;

	// 502F88(mBGAArray)
	*reinterpret_cast<PINT32>(0x0041BDDA) = 0x534;

	*reinterpret_cast<PINT32>(0x00481036) = 0x534;
	*reinterpret_cast<PINT32>(0x00481237) = 0x534;
	*reinterpret_cast<PINT32>(0x004812C8) = 0x534;
	*reinterpret_cast<PINT32>(0x004814FF) = 0x534;
	*reinterpret_cast<PINT32>(0x0048170C) = 0x534;
	*reinterpret_cast<PINT32>(0x0048179D) = 0x534;
	*reinterpret_cast<PINT32>(0x00481A7D) = 0x534;
	*reinterpret_cast<PINT32>(0x00481B72) = 0x534;
	*reinterpret_cast<PINT32>(0x00481CE3) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA293) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA2DB) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA704) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA757) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA926) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA979) = 0x534;
	*reinterpret_cast<PINT32>(0x004BAA18) = 0x534;
	*reinterpret_cast<PINT32>(0x004CA593) = 0x534;

	// 502FA4(mBGAPathArray)
	*reinterpret_cast<PINT32>(0x0041BE06) = 0x534;

	*reinterpret_cast<PINT32>(0x004B4686) = 0x534;
	*reinterpret_cast<PINT32>(0x004B4755) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA64A) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA6AD) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA7FB) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA8A1) = 0x534;
	*reinterpret_cast<PINT32>(0x004BAA45) = 0x534;

	// 502FC0(mBGADescArray)
	*reinterpret_cast<PINT32>(0x0041BE32) = 0x534;

	*reinterpret_cast<PINT32>(0x004810D6) = 0x534;
	*reinterpret_cast<PINT32>(0x0048129E) = 0x534;
	*reinterpret_cast<PINT32>(0x004812F1) = 0x534;
	*reinterpret_cast<PINT32>(0x004815AB) = 0x534;
	*reinterpret_cast<PINT32>(0x00481773) = 0x534;
	*reinterpret_cast<PINT32>(0x004817C6) = 0x534;
	*reinterpret_cast<PINT32>(0x00481B49) = 0x534;
	*reinterpret_cast<PINT32>(0x00481B9C) = 0x534;
	*reinterpret_cast<PINT32>(0x004BA9EB) = 0x534;
	*reinterpret_cast<PINT32>(0x004BAC61) = 0x534;
	*reinterpret_cast<PINT32>(0x004BACC4) = 0x534;

	// Ƃ݂
	WriteCall32(0x004DEB8B, LoadSoundHook);

	// ҂񂮂݂
	WriteCall32(0x004C8554, LoadPictureHook);
	WriteJump32(0x004C7A30, CreateSurfaceFromFileCustomHook);

	// ҂񂮂݂݂܂
	WritePush32(0x0049F2D5, 0x0049F2E6);
	WriteJump32(0x0049F2DA, CreateSurfaceFromFileCustomHook2);

	WritePush32(0x0049F368, 0x0049F37A);
	WriteJump32(0x0049F36D, CreateSurfaceFromFileCustomHook2);

	WritePush32(0x0049F65E, 0x0049F66F);
	WriteJump32(0x0049F663, CreateSurfaceFromFileCustomHook2);

	WritePush32(0x0049F710, 0x0049F721);
	WriteJump32(0x0049F715, CreateSurfaceFromFileCustomHook2);

	// ǂݍ݋
	WritePush32(0x004DECB1, reinterpret_cast<DWORD_PTR>(DSBuff_LoadSoundEpilogueHook));

	CProfileIO profile("\\_nbmfiles\\nbm_conf.ini");

	//if(profile.ReadInteger("Glasopal", "CustomOptions", 0)) {
	//	// Ղ傻
	//	WriteCall32(0x004B8FFA, RelocateObjHook);
	//	WriteCall32(0x004B91F8, RelocateObjHook);

	//	// 炴[Ղ傻
	//	*reinterpret_cast<PUINT32 *>(0x004B0395) = 0xFE30B5FF;
	//	*reinterpret_cast<PUINT16 *>(0x004B0399) = 0xFFFF;
	//	WriteCall32(0x004B039B, SaratherHook);

	//	// ؂邨Ղ񂠂ځ[
	//	WriteCall32(0x00453AD6, ShowResultSceneHook);
	//	WriteCall32(0x00493D1C, ShowResultSceneHook);
	//	WriteCall32(0x00453AFF, WriteToNSDHook);
	//	WriteCall32(0x00493DAE, WriteToNSDHook);
	//}

	if(profile.ReadInteger("Glasopal", "CustomSpecifiedBMSPlay", 0)) {
		// Ă傤
		WriteCall32(0x00453970, SimpleLoadBMS_Hook);
	}

	if(profile.ReadInteger("Glasopal", "DisableIME", 0)) {
		// Ղꂢ[
		WriteCall32(0x004539AD, PlayLoopHook);
		WriteCall32(0x004938FA, PlayLoopHook);
	}

	if(profile.ReadInteger("Glasopal", "BGAMovie", 0)) {
		// ނ
		WriteCall32(0x0049399C, UnloadBMSHook);
		WriteCall32(0x004539B9, UnloadBMSHook);
		WriteCall32(0x004BAA93, LoadBGAHook_Stub);
		WriteCall32(0x004706AF, RenderBGAHook);
		WritePush32(0x00484652, reinterpret_cast<DWORD_PTR>(SubclassWndProcHook));
	}

	if(profile.ReadInteger("Glasopal", "WindowMemorize", 0)) {
		// 
		WriteCall32(0x0045A56B, InitMainFrmHook);
	}

	if(profile.ReadInteger("Glasopal", "MinimumComboEffect", 0)) {
		// ӂӂ炵Ȃ
		//	xor eax,eax
		//	mov dword ptr [ebp-8ch],eax
		//	mov dword ptr [ebp-88h],eax
		//	jmp short 0047B1FBh
		*reinterpret_cast<PUINT32>(0x0047B192) = 0x8589C033;
		*reinterpret_cast<PUINT32>(0x0047B196) = 0xFFFFFF74;
		*reinterpret_cast<PUINT32>(0x0047B19A) = 0xFF788589;
		*reinterpret_cast<PUINT32>(0x0047B19E) = 0x59EBFFFF;
		// ƂłȂ
		WriteJump32(0x0043E21C, reinterpret_cast<void *>(0x0043E364));
		WriteJump32(0x0043E783, reinterpret_cast<void *>(0x0043E8CA));
	}

	//if(profile.ReadInteger("Glasopal", "CustomReplay", 0)) {
	//	WriteCall32(0x004B8FFA, BMSRelocateObjHook);
	//	WriteCall32(0x004B91F8, BMSRelocateObjHook);
	//	WriteCall32(0x0043B6C7, __vbaFileCloseHook1);
	//	*reinterpret_cast<BYTE *>(0x0043B6CC) = 0x90;
	//	WriteCall32(0x0043B32A, __vbaFileCloseHook2);
	//	*reinterpret_cast<BYTE *>(0x0043B32F) = 0x90; */
	//}

	//// [邠ځ[
	//DWORD oldpr;
	//VirtualProtect(reinterpret_cast<LPVOID>(0x004E4B6E), 8, PAGE_READWRITE, &oldpr);
	//*reinterpret_cast<LPDWORD>(0x004E4B6E) = 0xFFFFFFB8;
	//*reinterpret_cast<LPDWORD>(0x004E4B72) = 0x9017EBFF;
	//VirtualProtect(reinterpret_cast<LPVOID>(0x004E4B6E), 8, oldpr, &oldpr);

	//// 낭ځ[
	//WriteMem<BYTE>(0x004BBE40, 0xc3);

	FlushInstructionCache(GetModuleHandle(NULL), NULL, 0);

	if(profile.ReadInteger("Glasopal", "AutoPriority", 0))
		SetPriorityClass(GetCurrentProcess(), ABOVE_NORMAL_PRIORITY_CLASS);

	reinterpret_cast<void (*)()>(0x40F538)();
}

void Hook()
{
	DWORD oldattr;

	*reinterpret_cast<UINT32 *>(0x0050CCFD) = reinterpret_cast<DWORD_PTR>(OnMain) - 0x0050CD01;
	*reinterpret_cast<UINT32 *>(0x0050D8BF) = 0x4256534D;
	*reinterpret_cast<UINT32 *>(0x0050D8C3) = 0x30364D56;
	*reinterpret_cast<UINT32 *>(0x0050D8C7) = 0x4C4C442E;

	FlushInstructionCache(GetModuleHandle(NULL), NULL, 0);
}

int APIENTRY DllMain(HMODULE hLibModule, ULONG ulReasonForCall, LPVOID lpReserved)
{
	switch(ulReasonForCall) {
	case DLL_PROCESS_ATTACH:
		g_hLibModule = hLibModule;
		DisableThreadLibraryCalls(hLibModule);
		Hook();
	}
	return TRUE;
}

extern "C"
__declspec(dllexport)
char DummyMessage[] = 
/*
 0123456789012345 */
"_CfXSO"
"@@^_C}"
"^@TWE";
